@@ -20,27 +20,30 @@
 # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
+from __future__ import annotations
+
 import ast
+import base64
+import contextlib
+import json
 import os
 import pprint
 import site
 import sys
-from collections.abc import Mapping
+import warnings
+from collections.abc import Mapping, Sequence
 from contextlib import nullcontext
 from copy import deepcopy
+from typing import Any, Literal, MutableMapping
+
+import yaml
 
 from ._lock import SerializableLock
 
-try:
-    import yaml
-except ImportError:
-    yaml = None
-
-
 no_default = "__no_default__"
 
 
-def canonical_name(k, config):
+def canonical_name(k: str, config: Mapping[str, Any]) -> str:
     """Return the canonical name for a key.
 
     Handles user choice of '-' or '_' conventions by standardizing on whichever
@@ -63,7 +66,12 @@ def canonical_name(k, config):
     return k
 
 
-def update(old, new, priority="new"):
+def update(
+    old: MutableMapping[str, Any],
+    new: Mapping[str, Any],
+    priority: Literal["old", "new", "new-defaults"] = "new",
+    defaults: Mapping | None = None,
+) -> Mapping[str, Any]:
     """Update a nested dictionary with values from another
 
     This is like dict.update except that it smoothly merges nested values
@@ -72,9 +80,12 @@ def update(old, new, priority="new"):
 
     Parameters
     ----------
-    priority: string {'old', 'new'}
+    priority: string {'old', 'new', 'new-defaults'}
         If new (default) then the new dictionary has preference.
         Otherwise the old dictionary does.
+        If 'new-defaults', a mapping should be given of the current defaults.
+        Only if a value in ``old`` matches the current default, it will be
+        updated with ``new``.
 
     Examples
     --------
@@ -88,6 +99,12 @@ def update(old, new, priority="new"):
     >>> update(a, b, priority='old')  # doctest: +SKIP
     {'x': 1, 'y': {'a': 2, 'b': 3}}
 
+    >>> d = {'x': 0, 'y': {'a': 2}}
+    >>> a = {'x': 1, 'y': {'a': 2}}
+    >>> b = {'x': 2, 'y': {'a': 3, 'b': 3}}
+    >>> update(a, b, priority='new-defaults', defaults=d)  # doctest: +SKIP
+    {'x': 1, 'y': {'a': 3, 'b': 3}}
+
     See Also
     --------
     donfig.config_obj.merge
@@ -99,15 +116,24 @@ def update(old, new, priority="new"):
         if isinstance(v, Mapping):
             if k not in old or old[k] is None:
                 old[k] = {}
-            update(old[k], v, priority=priority)
+            update(
+                old[k],
+                v,
+                priority=priority,
+                defaults=defaults.get(k) if defaults else None,
+            )
         else:
-            if priority == "new" or k not in old:
+            if (
+                priority == "new"
+                or k not in old
+                or (priority == "new-defaults" and defaults and k in defaults and defaults[k] == old[k])
+            ):
                 old[k] = v
 
     return old
 
 
-def merge(*dicts):
+def merge(*dicts: Mapping) -> dict:
     """Update a sequence of nested dictionaries
 
     This prefers the values in the latter dictionaries to those in the former
@@ -124,13 +150,13 @@ def merge(*dicts):
     donfig.config_obj.update
 
     """
-    result = {}
+    result: dict = {}
     for d in dicts:
         update(result, d)
     return result
 
 
-def collect_yaml(paths):
+def collect_yaml(paths: Sequence[str]) -> list[dict]:
     """Collect configuration from yaml files
 
     This searches through a list of paths, expands to find all yaml or json
@@ -147,8 +173,7 @@ def collect_yaml(paths):
                         sorted(
                             os.path.join(path, p)
                             for p in os.listdir(path)
-                            if os.path.splitext(p)[1].lower()
-                            in (".json", ".yaml", ".yml")
+                            if os.path.splitext(p)[1].lower() in (".json", ".yaml", ".yml")
                         )
                     )
                 except OSError:
@@ -161,18 +186,33 @@ def collect_yaml(paths):
 
     # Parse yaml files
     for path in file_paths:
-        try:
-            with open(path) as f:
-                data = yaml.safe_load(f.read()) or {}
-                configs.append(data)
-        except OSError:
-            # Ignore permission errors
-            pass
+        config = _load_config_file(path)
+        if config is not None:
+            configs.append(config)
 
     return configs
 
 
-def collect_env(prefix, env=None):
+def _load_config_file(path: str) -> dict | None:
+    try:
+        with open(path) as f:
+            config = yaml.safe_load(f.read())
+    except OSError:
+        # Ignore permission errors
+        return None
+    except Exception as exc:
+        raise ValueError(f"A config file at {path!r} is malformed, original error " f"message:\n\n{exc}") from None
+    if config is not None and not isinstance(config, dict):
+        raise ValueError(
+            f"A config file at {path!r} is malformed - config files must have "
+            f"a dict as the top level object, got a {type(config).__name__} instead"
+        )
+    return config
+
+
+def collect_env(
+    prefix: str, env: Mapping[str, str] | None = None, deprecations: MutableMapping[str, str | None] | None = None
+) -> dict:
     """Collect config from environment variables
 
     This grabs environment variables of the form "DASK_FOO__BAR_BAZ=123" and
@@ -186,7 +226,13 @@ def collect_env(prefix, env=None):
     """
     if env is None:
         env = os.environ
-    d = {}
+
+    serial_env = f"{prefix}_INTERNAL_INHERIT_CONFIG"
+    if serial_env in env:
+        d = deserialize(env[serial_env])
+    else:
+        d = {}
+
     prefix_len = len(prefix)
     for name, value in env.items():
         if name.startswith(prefix):
@@ -196,10 +242,10 @@ def collect_env(prefix, env=None):
             except (SyntaxError, ValueError):
                 d[varname] = value
 
-    result = {}
+    result: dict = {}
     # fake thread lock to use set functionality
     lock = nullcontext()
-    ConfigSet(result, lock, d)
+    ConfigSet(result, lock, deprecations or {}, d)
     return result
 
 
@@ -223,17 +269,53 @@ class ConfigSet:
 
     """
 
-    def __init__(self, config, lock, arg=None, **kwargs):
+    def __init__(
+        self,
+        config: MutableMapping,
+        lock: SerializableLock | contextlib.AbstractContextManager,
+        deprecations: MutableMapping[str, str | None],
+        arg: Mapping | None = None,
+        **kwargs,
+    ):
         with lock:
             self.config = config
-            self._record = []
+            self.deprecations = deprecations
+            self._record: list[tuple[str, tuple, Any]] = []
 
             if arg is not None:
                 for key, value in arg.items():
+                    key = self._check_deprecations(key)
                     self._assign(key.split("."), value, config)
             if kwargs:
                 for key, value in kwargs.items():
-                    self._assign(key.split("__"), value, config)
+                    key = key.replace("__", ".")
+                    key = self._check_deprecations(key)
+                    self._assign(key.split("."), value, config)
+
+    def _check_deprecations(self, key: str):
+        """Check if the provided value has been renamed or removed.
+
+        Parameters
+        ----------
+        key : str
+            The configuration key to check
+
+        Returns
+        -------
+        new: str
+            The proper key, whether the original (if no deprecation) or the aliased
+            value
+
+        """
+        if key in self.deprecations:
+            new = self.deprecations[key]
+            if new:
+                warnings.warn(f"Configuration key {key!r} has been deprecated. " f"Please use {new!r} instead")
+                return new
+            else:
+                raise ValueError(f"Configuration value {key!r} has been removed")
+        else:
+            return key
 
     def __enter__(self):
         return self.config
@@ -254,7 +336,14 @@ class ConfigSet:
                 else:
                     d.pop(path[-1], None)
 
-    def _assign(self, keys, value, d, path=(), record=True):
+    def _assign(
+        self,
+        keys: Sequence[str],
+        value: Any,
+        d: MutableMapping,
+        path: tuple[str, ...] = (),
+        record: bool = True,
+    ) -> None:
         """Assign value into a nested configuration dictionary
 
         Parameters
@@ -265,7 +354,7 @@ class ConfigSet:
         d : dict
             The part of the nested dictionary into which we want to assign the
             value
-        path : List[str]
+        path : tuple[str], optional
             Used internally to hold the path of old values
         record : bool, optional
             Whether this operation needs to be recorded to allow for rollback.
@@ -325,13 +414,14 @@ def expand_environment_variables(config):
 class Config:
     def __init__(
         self,
-        name,
-        defaults=None,
-        paths=None,
-        env=None,
-        env_var=None,
-        root_env_var=None,
-        env_prefix=None,
+        name: str,
+        defaults: list[Mapping[str, Any]] | None = None,
+        paths: list[str] | None = None,
+        env: Mapping[str, str] | None = None,
+        env_var: str | None = None,
+        root_env_var: str | None = None,
+        env_prefix: str | None = None,
+        deprecations: Mapping[str, str | None] | None = None,
     ):
         if root_env_var is None:
             root_env_var = f"{name.upper()}_ROOT_CONFIG"
@@ -354,6 +444,8 @@ class Config:
             paths.append(main_path)
         else:
             main_path = os.path.join(os.path.expanduser("~"), ".config", name)
+        if deprecations is None:
+            deprecations = {}
 
         # Remove duplicate paths while preserving ordering
         paths = list(reversed(list(dict.fromkeys(reversed(paths)))))
@@ -364,7 +456,9 @@ class Config:
         self.main_path = main_path
         self.paths = paths
         self.defaults = defaults or []
-        self.config = {}
+        self.deprecations = deprecations
+
+        self.config: MutableMapping[str, Any] = {}
         self.config_lock = SerializableLock()
         self.refresh()
 
@@ -381,16 +475,16 @@ class Config:
     def pprint(self, **kwargs):
         return pprint.pprint(self.config, **kwargs)
 
-    def collect(self, paths=None, env=None):
+    def collect(self, paths: list[str] | None = None, env: Mapping[str, str] | None = None) -> dict:
         """Collect configuration from paths and environment variables
 
         Parameters
         ----------
-        paths : List[str]
+        paths : list[str]
             A list of paths to search for yaml config files. Defaults to the
             paths passed when creating this object.
 
-        env : dict
+        env : Mapping[str, str]
             The system environment variables to search through. Defaults to
             the environment dictionary passed when creating this object.
 
@@ -416,7 +510,7 @@ class Config:
 
         return merge(*configs)
 
-    def refresh(self, **kwargs):
+    def refresh(self, **kwargs) -> None:
         """Update configuration by re-reading yaml files and env variables.
 
         This goes through the following stages:
@@ -444,7 +538,7 @@ class Config:
 
         update(self.config, self.collect(**kwargs))
 
-    def get(self, key, default=no_default):
+    def get(self, key: str, default: Any = no_default) -> Any:
         """Get elements from global config
 
         Use '.' for nested access
@@ -480,18 +574,20 @@ class Config:
                     raise
         return result
 
-    def update_defaults(self, new):
+    def update_defaults(self, new: Mapping) -> None:
         """Add a new set of defaults to the configuration
 
         It does two things:
 
         1.  Add the defaults to a collection to be used by refresh() later
-        2.  Updates the config with the new configuration
-            prioritizing older values over newer ones
+        2.  Updates the global config with the new configuration.
+            Old values are prioritized over new ones, unless the current value
+            is the old default, in which case it's updated to the new default.
 
         """
+        current_defaults = merge(*self.defaults)
         self.defaults.append(new)
-        update(self.config, new, priority="old")
+        update(self.config, new, priority="new-defaults", defaults=current_defaults)
 
     def to_dict(self):
         """Return dictionary copy of configuration.
@@ -525,7 +621,7 @@ class Config:
         """
         self.config = update(self.config, new, priority=priority)
 
-    def expand_environment_variables(self):
+    def expand_environment_variables(self) -> None:
         """Expand any environment variables in this configuration in-place.
 
         See :func:`~donfig.config_obj.expand_environment_variables` for more information.
@@ -533,7 +629,7 @@ class Config:
         """
         self.config = expand_environment_variables(self.config)
 
-    def rename(self, aliases):
+    def rename(self, aliases: Mapping) -> None:
         """Rename old keys to new keys
 
         This helps migrate older configuration versions over time
@@ -589,9 +685,9 @@ class Config:
         donfig.Config.get
 
         """
-        return ConfigSet(self.config, self.config_lock, arg=arg, **kwargs)
+        return ConfigSet(self.config, self.config_lock, self.deprecations, arg=arg, **kwargs)
 
-    def ensure_file(self, source, destination=None, comment=True):
+    def ensure_file(self, source: str, destination: str | None = None, comment: bool = True) -> None:
         """Copy file to default location if it does not already exist
 
         This tries to move a default configuration file to a default location if
@@ -636,12 +732,7 @@ class Config:
                     lines = list(f)
 
                 if comment:
-                    lines = [
-                        "# " + line
-                        if line.strip() and not line.startswith("#")
-                        else line
-                        for line in lines
-                    ]
+                    lines = ["# " + line if line.strip() and not line.startswith("#") else line for line in lines]
 
                 with open(tmp, "w") as f:
                     f.write("".join(lines))
@@ -652,3 +743,52 @@ class Config:
                     os.remove(tmp)
         except OSError:
             pass
+
+    def serialize(self) -> str:
+        """Serialize conifg data into a string.
+
+        See :func:`serialize` for more information.
+
+        """
+        return serialize(self.config)
+
+
+def serialize(data: Any) -> str:
+    """Serialize config data into a string.
+
+    Typically used to pass config via the ``MYPKG_INTERNAL_INHERIT_CONFIG`` environment variable.
+
+    Parameters
+    ----------
+    data: json-serializable object
+        The data to serialize
+
+    Returns
+    -------
+    serialized_data: str
+        The serialized data as a string
+
+    """
+    return base64.urlsafe_b64encode(json.dumps(data).encode()).decode()
+
+
+def deserialize(data: str) -> Any:
+    """De-serialize config data into the original object.
+
+    Typically used when receiving config via the
+    ``MYPKG_INTERNAL_INHERIT_CONFIG`` environment variable. This is
+    automatically called when a :class:`Config` is created and environment
+    variables are loaded.
+
+    Parameters
+    ----------
+    data: str
+        String serialized by :func:`donfig.serialize`
+
+    Returns
+    -------
+    deserialized_data: obj
+        The de-serialized data
+
+    """
+    return json.loads(base64.urlsafe_b64decode(data.encode()).decode())
